@@ -1,5 +1,4 @@
 # pylint:disable=line-too-long
-
 """Parser for ssh public keys. Currently supports ssh-rsa, ssh-dsa, ssh-ed25519 and ssh-dss keys.
 
 import sys
@@ -14,27 +13,28 @@ except InvalidKeyError:
     sys.exit(1)
 print(ssh_key.bits)"""
 
+from .exceptions import (
+    InvalidKeyError, InvalidKeyLengthError, InvalidOptionNameError, InvalidOptionsError, InvalidTypeError,
+    MalformedDataError, MissingMandatoryOptionValueError, TooLongKeyError, TooShortKeyError, UnknownOptionNameError
+)
+from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives.asymmetric.dsa import DSAParameterNumbers, DSAPublicNumbers
+from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
+from urllib.parse import urlparse
+
 import base64
 import binascii
+import ecdsa
 import hashlib
 import re
 import struct
 import sys
 import warnings
 
-import ecdsa
-from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives.asymmetric.dsa import DSAParameterNumbers, DSAPublicNumbers
-from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
-
-from .exceptions import (InvalidKeyError, InvalidKeyLengthError, InvalidOptionNameError, InvalidOptionsError,
-                         InvalidTypeError, MalformedDataError, MissingMandatoryOptionValueError, TooLongKeyError,
-                         TooShortKeyError, UnknownOptionNameError)
-
 __all__ = ["AuthorizedKeysFile", "SSHKey"]
 
 
-class AuthorizedKeysFile(object):  # pylint:disable=too-few-public-methods
+class AuthorizedKeysFile:  # pylint:disable=too-few-public-methods
     """Represents a full authorized_keys file.
 
     Comments and empty lines are ignored."""
@@ -55,7 +55,7 @@ class AuthorizedKeysFile(object):  # pylint:disable=too-few-public-methods
             self.keys.append(ssh_key)
 
 
-class SSHKey(object):  # pylint:disable=too-many-instance-attributes
+class SSHKey:  # pylint:disable=too-many-instance-attributes
     """Represents a single SSH keypair.
 
     ssh_key = SSHKey(key_data, strict=True)
@@ -85,23 +85,26 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
     RSA_MIN_LENGTH_LOOSE = 768
     RSA_MAX_LENGTH_LOOSE = 16384
 
-    # Valid as of OpenSSH_6.9p1
+    # Valid as of OpenSSH_8.3
     # argument name, value is mandatory. Options are case-insensitive, but this list must be in lowercase.
     OPTIONS_SPEC = [
         ("agent-forwarding", False),
         ("cert-authority", False),
         ("command", True),
         ("environment", True),
+        ("expiry-time", True),
         ("from", True),
         ("no-agent-forwarding", False),
         ("no-port-forwarding", False),
         ("no-pty", False),
         ("no-user-rc", False),
         ("no-x11-forwarding", False),
+        ("permitlisten", True),
         ("permitopen", True),
         ("port-forwarding", False),
         ("principals", True),
         ("pty", False),
+        ("no-touch-required", False),
         ("restrict", False),
         ("tunnel", True),
         ("user-rc", False),
@@ -134,7 +137,7 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
                 pass
 
     def __str__(self):
-        return "Key type: %s, bits: %s, options: %s" % (self.key_type, self.bits, self.options)
+        return "Key type: %s, bits: %s, options: %s" % (self.key_type.decode(), self.bits, self.options)
 
     def reset(self):
         """Reset all data fields."""
@@ -172,15 +175,17 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
         # Unpack length of data field
         try:
             requested_data_length = struct.unpack('>I', data[current_position:current_position + self.INT_LEN])[0]
-        except struct.error:
-            raise MalformedDataError("Unable to unpack %s bytes from the data" % self.INT_LEN)
+        except struct.error as ex:
+            raise MalformedDataError("Unable to unpack %s bytes from the data" % self.INT_LEN) from ex
 
         # Move pointer to the beginning of the data field
         current_position += self.INT_LEN
         remaining_data_length = len(data[current_position:])
 
         if remaining_data_length < requested_data_length:
-            raise MalformedDataError("Requested %s bytes, but only %s bytes available." % (requested_data_length, remaining_data_length))
+            raise MalformedDataError(
+                "Requested %s bytes, but only %s bytes available." % (requested_data_length, remaining_data_length)
+            )
 
         next_data = data[current_position:current_position + requested_data_length]
         # Move pointer to the end of the data field
@@ -204,7 +209,7 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
     def _split_key(self, data):
         options_raw = None
         # Terribly inefficient way to remove options, but hey, it works.
-        if not data.startswith("ssh-") and not data.startswith("ecdsa-"):
+        if not data.startswith("ssh-") and not data.startswith("ecdsa-") and not data.startswith("sk-"):
             quote_open = False
             for i, character in enumerate(data):
                 if character == '"':  # only double quotes are allowed, no need to care about single quotes
@@ -240,8 +245,8 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
         """Decode base64 coded part of the key."""
         try:
             decoded_key = base64.b64decode(pubkey_content.encode("ascii"))
-        except (TypeError, binascii.Error):
-            raise MalformedDataError("Unable to decode the key")
+        except (TypeError, binascii.Error) as ex:
+            raise MalformedDataError("Unable to decode the key") from ex
         return decoded_key
 
     @classmethod
@@ -262,15 +267,15 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
                 opt_name = opt
                 opt_value = True
             if " " in opt_name or not self.OPTION_NAME_RE.match(opt_name):
-                raise InvalidOptionNameError("%s is not valid option name." % opt_name)
+                raise InvalidOptionNameError("%s is not a valid option name." % opt_name)
             if self.strict_mode:
                 for valid_opt_name, value_required in self.OPTIONS_SPEC:
                     if opt_name.lower() == valid_opt_name:
                         if value_required and opt_value is True:
-                            raise MissingMandatoryOptionValueError("%s is missing mandatory value." % opt_name)
+                            raise MissingMandatoryOptionValueError("%s is missing a mandatory value." % opt_name)
                         break
                 else:
-                    raise UnknownOptionNameError("%s is unrecognized option name." % opt_name)
+                    raise UnknownOptionNameError("%s is an unrecognized option name." % opt_name)
             if opt_name not in parsed_options:
                 parsed_options[opt_name] = []
             parsed_options[opt_name].append(opt_value)
@@ -312,9 +317,13 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
             min_length = self.RSA_MIN_LENGTH_LOOSE
             max_length = self.RSA_MAX_LENGTH_LOOSE
         if self.bits < min_length:
-            raise TooShortKeyError("%s key data can not be shorter than %s bits (was %s)" % (self.key_type, min_length, self.bits))
+            raise TooShortKeyError(
+                "%s key data can not be shorter than %s bits (was %s)" % (self.key_type.decode(), min_length, self.bits)
+            )
         if self.bits > max_length:
-            raise TooLongKeyError("%s key data can not be longer than %s bits (was %s)" % (self.key_type, max_length, self.bits))
+            raise TooLongKeyError(
+                "%s key data can not be longer than %s bits (was %s)" % (self.key_type.decode(), max_length, self.bits)
+            )
         return current_position
 
     def _process_ssh_dss(self, data):
@@ -336,9 +345,13 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
             min_length = self.DSA_MIN_LENGTH_LOOSE
             max_length = self.DSA_MAX_LENGTH_LOOSE
         if p_bits < min_length:
-            raise TooShortKeyError("%s key can not be shorter than %s bits (was %s)" % (self.key_type, min_length, p_bits))
+            raise TooShortKeyError(
+                "%s key can not be shorter than %s bits (was %s)" % (self.key_type.decode(), min_length, p_bits)
+            )
         if p_bits > max_length:
-            raise TooLongKeyError("%s key data can not be longer than %s bits (was %s)" % (self.key_type, max_length, p_bits))
+            raise TooLongKeyError(
+                "%s key data can not be longer than %s bits (was %s)" % (self.key_type.decode(), max_length, p_bits)
+            )
 
         dsa_parameters = DSAParameterNumbers(data_fields["p"], data_fields["q"], data_fields["g"])
         self.dsa = DSAPublicNumbers(data_fields["y"], dsa_parameters).public_key(default_backend())
@@ -357,8 +370,8 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
         try:
             # data starts with \x04, which should be discarded.
             ecdsa_key = ecdsa.VerifyingKey.from_string(key_data[1:], curve, hash_algorithm)
-        except AssertionError:
-            raise InvalidKeyError("Invalid ecdsa key")
+        except AssertionError as ex:
+            raise InvalidKeyError("Invalid ecdsa key") from ex
         self.bits = int(curve_information.replace(b"nistp", b""))
         self.ecdsa = ecdsa_key
         return current_position
@@ -382,17 +395,46 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
             raise InvalidKeyLengthError("ed25519 keys must be 256 bits (was %s bits)" % self.bits)
         return current_position
 
+    def _validate_application_string(self, application):
+        """Validates Application string.
+
+        Has to be an URL starting with "ssh:". See ssh-keygen(1)."""
+
+        try:
+            parsed_url = urlparse(application)
+        except ValueError as err:
+            raise InvalidKeyError(f"Application string: {err}") from err
+        if parsed_url.scheme != b"ssh":
+            raise InvalidKeyError('Application string must begin with "ssh:"')
+
+    def _process_sk_ecdsa_sha(self, data):
+        """Parses sk_ecdsa-sha public keys."""
+        current_position = self._process_ecdsa_sha(data)
+        current_position, application = self._unpack_by_int(data, current_position)
+        self._validate_application_string(application)
+        return current_position
+
+    def _process_sk_ed25519(self, data):
+        """Parses sk_ed25519 public keys."""
+        current_position = self._process_ed25516(data)
+        current_position, application = self._unpack_by_int(data, current_position)
+        self._validate_application_string(application)
+        return current_position
+
     def _process_key(self, data):
         if self.key_type == b"ssh-rsa":
             return self._process_ssh_rsa(data)
-        elif self.key_type == b"ssh-dss":
+        if self.key_type == b"ssh-dss":
             return self._process_ssh_dss(data)
-        elif self.key_type.strip().startswith(b"ecdsa-sha"):
+        if self.key_type.strip().startswith(b"ecdsa-sha"):
             return self._process_ecdsa_sha(data)
-        elif self.key_type == b"ssh-ed25519":
+        if self.key_type == b"ssh-ed25519":
             return self._process_ed25516(data)
-        else:
-            raise NotImplementedError("Invalid key type: %s" % self.key_type)
+        if self.key_type.strip().startswith(b"sk-ecdsa-sha"):
+            return self._process_sk_ecdsa_sha(data)
+        if self.key_type.strip().startswith(b"sk-ssh-ed25519"):
+            return self._process_sk_ed25519(data)
+        raise NotImplementedError("Invalid key type: %s" % self.key_type.decode())
 
     def parse(self, keydata=None):
         """Validates SSH public key.
@@ -426,7 +468,7 @@ class SSHKey(object):  # pylint:disable=too-many-instance-attributes
         # Check key type
         current_position, unpacked_key_type = self._unpack_by_int(self._decoded_key, 0)
         if key_type is not None and key_type != unpacked_key_type.decode():
-            raise InvalidTypeError("Keytype mismatch: %s != %s" % (key_type, unpacked_key_type))
+            raise InvalidTypeError("Keytype mismatch: %s != %s" % (key_type, unpacked_key_type.decode()))
 
         self.key_type = unpacked_key_type
 
